include "logic.idr" ;
include "preorder.idr" ;

namespace Pareto {
params (nA' : Nat , nG' : Nat) {

  nA : Nat   ;
  nA = S nA' ;
  nG : Nat   ;
  nG = S nG' ;
  
  params (Feasible : Vect (Vect Float nG) nA -> Set ,
          prefs    : Fin nA -> TotalPreorder (Vect (Vect Float nG) nA) ) {

---------------------------------------------------------------------------------
--                                 DEFINITIONS                                 --
--------------------------------------------------------------------------------- 

    WeakPareto : Vect (Vect Float nG) nA -> Set ;
    WeakPareto xss = And (Feasible xss) 
                         (forall (xss' : Vect (Vect Float nG) nA) ->
                                 (forall (i : Fin nA) -> gt (prefs i) xss' xss) ->
                                 Not (Feasible xss')) ;

    StrongPareto : Vect (Vect Float nG) nA -> Set;
    StrongPareto xss = And (Feasible xss)
                           (forall (xss' : Vect (Vect Float nG) nA) ->
                            (And (forall (i : Fin nA) -> gte (prefs i) xss' xss)
                                 (Sigma (Fin nA) (\ i =>
                                                 gt (prefs i) xss' xss))) ->
                            Not (Feasible xss')) ;


---------------------------------------------------------------------------------
--                      STRONG PARETO IMPLIES WEAK PARETO                      --
---------------------------------------------------------------------------------

    strongimplweak : (xss : Vect (Vect Float nG) nA) -> 
                     StrongPareto xss -> WeakPareto xss ;

---------------------------------------------------------------------------------

    strongimplweak xss (and_intro fxss sxss) = 
       and_intro fxss 
       (\ xss' , gtxss'xss , fxss' => 
               sxss xss' (and_intro ? ?swp2) fxss'); -- proofs in Appendix
  }

}

}

------------------------------------------------------------------------
--                              APPENDIX                              --
------------------------------------------------------------------------
strongimplweak_1 proof {
        %intro;
        %refine strictimplweak;
        %intro L;
        %fill gtxss'xss X L;
        %qed;
};

swp2 proof {
	%intro;
	%refine Exists;
	%fill fO;
	%fill gtxss'xss fO;
	%qed;
};
