include "logic.idr" ;

data [noElim] IsPreorder : {X : Set} -> (lte : X -> X -> Set) -> Set where
  isPreorder : {X : Set} -> {lte : X -> X -> Set} ->
               (reflexive :  (x : X) -> lte x x) ->
               (transitive : (x : X) -> (x' : X) -> (x'' : X) ->
                             lte x x' -> lte x' x'' -> lte x x'') ->
               IsPreorder lte ;

data Preorder : (X : Set) -> Set where
  mkPreorder : (lte : X -> X -> Set) ->
               IsPreorder lte ->
               Preorder X ;

---------------------------------------------------------------------------------
--                 PREORDER EXTRACTORS                                         --
---------------------------------------------------------------------------------
lte' : Preorder X -> (X -> X -> Set) ;
lte' (mkPreorder f p) = f ;
refl' : (p : Preorder X) -> ((x : X) -> lte' p x x);
tran' : (p : Preorder X) -> ((x : X) -> (x' : X) -> (x'' : X) ->
                      (lte' p x x') -> (lte' p x' x'') -> (lte' p x x''));
---------------------------------------------------------------------------------

data TotalPreorder : (X : Set) -> Set where
  mkTotalPreorder : (pr : Preorder X) ->
                    (total : (x : X) -> (x' : X) -> 
                             Or (lte' pr x x') (lte' pr x' x)) ->
                    TotalPreorder X ;

                
---------------------------------------------------------------------------------
--           TOTAL PREORDER EXTRACTORS                                         --
---------------------------------------------------------------------------------
lte : TotalPreorder X -> X -> X -> Set ;
lte (mkTotalPreorder pr t) = lte' pr ;

reflexive  : (p : TotalPreorder X) -> ((x : X) -> lte p x x);
reflexive  (mkTotalPreorder pr total) = refl' pr;

transitive : (p : TotalPreorder X) -> ((x : X) -> (x' : X) -> (x'' : X) ->
                      (lte p x x') -> (lte p x' x'') -> (lte p x x''));
transitive (mkTotalPreorder pr total) = tran' pr;

totalityProof : (tp : TotalPreorder X) -> 
                ((x : X) -> (x' : X) -> 
                 Or (lte tp x x') (lte tp x' x)) ;
totalityProof (mkTotalPreorder pr total) = total ;

gt : TotalPreorder X -> X -> X -> Set ;
gt tp x x' = Not (lte tp x x') ;

gte : TotalPreorder X -> X -> X -> Set ;
gte tp x x' = lte tp x' x ;

---------------------------------------------------------------------------------

---------------------------------------------------------------------------------
--                    x > x' -> x >= x'                                        --
---------------------------------------------------------------------------------

strictimplweak : {X : Set} -> (tp : TotalPreorder X) -> (x : X) -> (x' : X) ->
                 gt tp x x' -> gte tp x x' ;
strictimplweak tp x x' gtxx' with totalityProof tp x x' {
                               |  or_intro_l y = ?siw_l ;
                               |  or_intro_r y = ?siw_r ;
}  

----------------------------------------------------------------
--                        APPENDIX                            --
----------------------------------------------------------------

siw_l proof {
        %intro X;
        %intro;
        %refine contradiction;
        %refine gtxx';
        %fill y;
        %qed;
};

siw_r proof {
        %intro X;
        %intro;
        %fill y;
        %qed;
};
