include "preorder.idr" ;
include "utils.idr" ;
include "pareto.idr";

params (nA' : Nat , nG' : Nat) {

  nA : Nat ;
  nA = S nA' ;
  nG : Nat ;
  nG = S nG' ;

  params (omega  : Vect (Vect Float nG) nA ,
          prices : Vect Float nG ,
          prefs  : Fin nA -> TotalPreorder (Vect Float nG)) {


---------------------- Walrasan Equilibrium --------------------

    Feasible : Vect (Vect Float nG) nA -> Set ;
    Feasible xss =  SumCols xss === SumCols omega ;

    Optimal : Vect (Vect Float nG) nA -> Set;
    Optimal xss = forall (i : Fin nA, xss' : Vect (Vect Float nG) nA) ->
                         gt (prefs i) (vlookup i xss') 
                                      (vlookup i xss) ->
                         gt floatOrder (prices .* (vlookup i xss'))
                                       (prices .* (vlookup i xss ));

    WalrasEq : Vect (Vect Float nG) nA -> Set;
    WalrasEq xss = And (Feasible xss) (Optimal xss);

---------------------------------------------------------------

    glte : Fin nA -> Vect (Vect Float nG) nA -> Vect (Vect Float nG) nA -> Set;
    glte i xss xss' = lte (prefs i) (vlookup i xss) (vlookup i xss');

    glteRefl : (i : Fin nA) -> (xss : Vect (Vect Float nG) nA) -> glte i xss xss;

    globalIsPreorder : (i : Fin nA) -> IsPreorder (glte i);
    globalIsPreorder i = isPreorder (glteRefl i) ?glteTran;

    globalPreorder : Fin nA -> Preorder (Vect (Vect Float nG) nA);
    globalPreorder i = mkPreorder (glte i) (globalIsPreorder i);

    globalPrefs : Fin nA -> TotalPreorder (Vect (Vect Float nG) nA);
    globalPrefs i = mkTotalPreorder (globalPreorder i) 
                                    (\ xss, xss' => totalityProof (prefs i) 
                                                                  (vlookup i xss) 
                                                                  (vlookup i xss')) ;

    sameCost : (xss : Vect (Vect Float nG) nA) -> Feasible xss ->
               ((prices .* (SumCols xss)) === (prices .* (SumCols omega)));

    mon : (xss : Vect (Vect Float nG) nA) -> (xss' : Vect (Vect Float nG) nA) ->
      ((i : Fin nA) -> 
       gt floatOrder (prices .* (vlookup i xss)) (prices .* (vlookup i xss'))) ->
       gt floatOrder (Sum (prices .*.* xss)) (Sum (prices .*.* xss'));

    distr : (xss : Vect (Vect Float nG) nA) ->
            ((Sum (prices .*.* xss)) === (prices .* (SumCols xss)));

    mon' : (xss : Vect (Vect Float nG) nA) -> (xss' : Vect (Vect Float nG) nA) ->
           ((i : Fin nA) -> gt floatOrder (prices .* (vlookup i xss))
                                          (prices .* (vlookup i xss'))) ->
           gt floatOrder (prices .* (SumCols xss)) (prices .* (SumCols xss'));
    mon' xss xss' gtpxipx'i = let dx   =  distr xss,
                                    dx'  =  distr xss',
                                    m    =  mon xss xss' gtpxipx'i
                               in ?;

    contra : gt floatOrder x y -> Not (x === y);

--------------------- First Welfare Theorem -------------------------------------

    firstWelfareTheorem : (xss : Vect (Vect Float nG) nA) ->
               WalrasEq xss -> WeakPareto nA' nG' Feasible globalPrefs xss;

---------------------------------------------------------------------------------
    firstWelfareTheorem xss (and_intro feasxss optimxss) =
     and_intro feasxss 
               (\ xss' , gtx'x, fx' => 
                contra (mon' xss' xss 
                            (\ i => optimxss i xss'  
                                ?gtproof)) ?)  ;   -- proof in Appendix

  }

}

------------------------------------------------------------------------
--                             APPENDIX                               --
------------------------------------------------------------------------
glteRefl proof {
        %intro;
        %compute;
        %refine reflexive;
        %qed;
};

glteTran proof {
        %intro;
        %fill transitive (prefs i) _ _ _ X2 X3;
        %qed;
};

sameCost proof {
        %intro;
        %rewrite __X_1;
        %refl;
        %qed;
};

mon'_1 proof {
        %intro nG', nA', xss', xss, prices, gtpxipx'i, prefs, omega, dx, dx';
        %rewrite <- dx;
        %rewrite <- dx';
        %intro m;
        %fill m;
        %qed;
};

gtproof proof {
        %intro;
        %fill gtx'x i X;
        %qed;
};

firstWelfareTheorem_1 proof {
        %intro;
        %rewrite <- feasxss;
        %rewrite <- fx';
        %refl;
        %qed;
};
